/*
	chiptune-simulator

	(c) Risto Holopainen 2012
	
	Copying and distribution of this file, with or without modification,
	are permitted in any medium without royalty provided the copyright
	notice and this notice are preserved.  This file is offered as-is,
	without any warranty.	

*/


#include "chip.h"
#include <sndfile.h>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <deque>

using std::deque;



struct note
{
	float frq;
	float amp;
	float dur;
};

float glissdetector(const deque <note> &);


void make_seq(deque <note> &SQ, int N)
{
	SQ.clear();
	float f, a = 0.95, d = 0.04;
	note T;
	T.amp = a;
	T.dur = d;
	
	for(int i=0; i<N; i++)
		{
		f = 165 + 55*i + 5*i*i - 55*(i%4);
		T.frq = f;
		SQ.push_back(T);
		}
}


void score_out(const deque <note> &DQ, int rep)
{

	static float st = 0;
	
	for(int k=0; k<rep; k++)
		{
		for(unsigned int n=0; n<DQ.size(); n++)
			{
			printf("i1 %.3f \t%.3f \t%.3f \t%.1f\n", 
				st, DQ[n].dur, DQ[n].amp, DQ[n].frq);
			st += DQ[n].dur;
			}
		}	

}


void muddify(deque <note> &DN)
{
	const unsigned int N = DN.size();
	float df, dr;

	for(unsigned int n=0; n<N-1; n++)
		{
		df = DN[n+1].frq - DN[n].frq; 
		dr = df/DN[n].frq - 1.;
		if( fabs(dr) < 0.20 && DN[n].frq > 1)
			{
			DN[n].frq -= 1;
			DN[n+1].frq *= 1.05;
			if(n > 1)
				{
				DN[n-1].dur += 0.05;
				DN[n].dur *= 0.87;
				}
			
			}
		else
			DN[n].frq += ((n&1) - 0.5) * 25;
		
		if(DN[n].dur < 0.01)	DN[n].dur += 0.02;
		if(DN[n].frq < 30)		DN[n].frq = 196 + n;
		if(DN[n].frq > 1760)	DN[n].frq *= 0.333;
		if(DN[n].amp < 0.06)	DN[n].amp = 1.0 - sqrt( (float) n/N );
		}
	printf("muddify: %d\n", DN.size());
}



void permutate(deque <note> &DQ, float G)
{
	const unsigned N = DQ.size();
	deque <note> TMP(N);
	
	for(unsigned int n=0; n<N; n++)
		{
		if(n & 1)
			TMP.push_back( DQ[n] );
		else
			TMP.push_front( DQ[n] );
		}
	
	for(unsigned int n=0; n<N; n++)
		{
		if((n & 1) && TMP[n].dur > 0.32)
			TMP[n].dur *= 0.4 *(1 + G);
		if(N < 100 || n != N/2)	// skip N/2 if N>100
			DQ[n] = TMP[n];
		}

	printf("permutate: %d\n", DQ.size());
}



void conk(deque <note> &DD)
{
	const int N = DD.size();
	note T;
	T.frq = 550 + 0.2*DD[0].frq; T.dur = 0.18 + 0.5*DD[0].dur; T.amp = 0.95;
	DD.push_front(T);
	if(N > 4 && DD[1].dur < 0.2) 
		{
		DD[1].dur *= 2;
		T.frq = 100 + 0.3* DD[N-3].frq;
		T.dur = 0.4;
		DD.push_back(T);
		DD[N-5].amp = 0.01;
		}
	printf("conk : %d \n", DD.size());
}


void sift(deque <note> &DD)
{
	const int N = DD.size();
	if (N > 6 && N < 20)
		{
		DD.pop_front();
		if(DD[1].frq > 60)
			DD[1].frq -= 45;
		}

	else if (N > 20)
		{
		DD.pop_back();
		DD[3].dur += 0.125;
		}

	if(N >= 132)
		{
		for(int n=1; n<N/2; n += 3)
			{
			DD[n].dur *= 0.71;
			if( DD[n].frq > 660)
				DD[n].amp = 600. / DD[n].frq;
			if( fabs(DD[n].frq - DD[n-1].frq) < 25 )
				DD.pop_front();
			}
		}
	
	printf("sift : %d\n", DD.size());
}


void havoc(deque <note> &HD)
{
	const int N = HD.size();
	
	note T;
	for(int j=0; j<N/3; j++)
		{
		T = HD[j];
		HD.push_back(T);
		HD.pop_front();		

		T.amp = 0.6 + .4*cos(24*T.amp);
		if(T.frq < 1800)
			T.frq *= 1.25 + .5*sin(6*T.dur);
		else
			{
			T.frq -= 500;
			T.amp = 1000./T.frq;
			}
		if(T.dur > 0.032)
			T.dur *= 0.6;
		else
			T.dur += 0.024;
		HD.push_back(T);
		}
	printf("havoc: %d\n", HD.size());
}


void shrimp(deque <note> &SD)
{
	const int N = SD.size();
	if(N > 25)
		for(int i=1; i<N/5; i++)
			{
			if(i&1 && SD[0].dur < 0.06)
				SD.pop_front();
			else
				SD.pop_back();
			}		
	
	if(N > 1 && SD[1].frq > 200)
		SD[1].frq *= 0.8;
	
	
	printf("shrimp %d\n", SD.size());
}


void tolchok(deque <note> &TN)
{
	const int N = TN.size();
	note X;
	deque <note> DQ;
	
	for(int j=0; j<N; j++)
		{
		X = TN[j];
		if(X.amp > 0.78 && X.frq < 900)
			DQ.push_front(X);
		else if(X.amp < 0.72 && X.dur < 0.06)
			{
			X.dur += 0.06;
			X.frq *= 0.95;
			X.amp += 0.24;
			DQ.push_back(X);
			}
		}

	if(TN.size() > 3 && DQ.size() > 3)
		TN.clear();

	int S = (DQ.size() > 6) ? 3 : 0;
	for(int n=S; n<DQ.size(); n++)
		TN.push_back(DQ[n]);


	printf("tolchok: %d\n", TN.size());
}


void score_event(const deque <note> &EQ, int k, FILE *sco, float st)
{
	// analyse eq --> 
	const int N = EQ.size();
	float dur = -0.05, amp = 0.64, frq = 0, sf = 0, rt;	// rt: score start-time relative.
	int instr = -1;
	amp += 0.15*sin(st);

	if(k < 5 && k != 2)	// + if etc.
		{
		instr = (k<2) ? 2 : 5;	// orgel--fagott, etc.

		for(int i=0; i<N; i++)
			{
			dur += EQ[i].dur;
			sf += EQ[i].frq;
			}
		float ff = sf/N;
		if(ff > 600) ff /= 2;
		if(ff > 415) ff /= 2;
		frq = 0.5*ff + N%24 -6;
		rt = (EQ[0].dur + EQ[N-1].dur)*sqrt(N);

		fprintf(sco, "i%d \t%.3f \t%.1f \t%.3f \t%.1f\n",
			instr, st, dur, .8*amp, frq);
		if(st > rt)
		fprintf(sco, "i%d \t%.3f \t%.1f \t%.3f \t%.1f\n",
			instr, st-rt, dur+rt, .8*amp, ff);
	}
	
	else if(k >= 5)	// analyse gliss
		{
		float t = st, u = st, skip = 0.12, gliss, ff, pan=0.5;
		gliss = glissdetector(EQ);
		if(gliss > 0.24) skip *= .84+gliss;
		instr = 1;
		for(int i=0; i<N; i++)
			{
			if(EQ[i].dur > 0.025 && t-u > skip && EQ[i].frq > 36 && EQ[i].frq < 660)
				{
				if(gliss > 0.40)
					ff = EQ[i].frq * 2 / ((i%2) + 3);
				else
					ff = EQ[i].frq / 2.;
				if(ff > 240)
					ff *= 0.8;

				u = t;	// hindrance for too swift repetitions
				fprintf(sco, "i%d \t%.3f \t%.3f \t%.3f \t%.1f \t%.3f\n",
					instr, t, EQ[i].dur+0.06, amp, ff, pan);
				if(EQ[i].dur > 0.06 && (i < 12 || i > 120) )	// extra note--echo
					{
					rt = EQ[i].dur/2 + 0.01;
					ff *= 1.2 - 0.5*amp;
					fprintf(sco, "i%d \t%.3f \t%.3f \t%.3f \t%.1f \t%.3f\n",
						instr, t+rt, EQ[i].dur-0.01, amp, ff, pan);
					pan = .95*(i&1);
					}
				else pan = (pan-0.5)*(pan-0.5)+0.45;
				}
			t += EQ[i].dur;
			}
		}
}


void store(const deque <note> &XD, deque <note> &YD)
{
	YD.clear();
	for(int n=0; n<XD.size(); n++)
		YD.push_back(XD[n]);
}


void combine(const deque <note> &XD, deque <note> &YC)
{
// copy XD --> YC (partially)

	const int N = XD.size();
	const int M = YC.size();
	
	if(M < 16 && N < 16)	// mirror
		{
		YC.clear();
		for(int i=0; i<N; i++)
			{
			YC.push_back(XD[i]);
			YC.push_front(XD[i]);
			YC[0].frq *= 2/3.;
			}
		return;
		}
	
	if( N < M )
		{
		float w;
		for(int i=1; i<N; i++)
			{
			w = (float) i/N;
			YC[i].amp = .5*XD[i].amp*w*(1-w) ;
			YC[i].dur = .5*(XD[i].dur + YC[i].dur);
			YC[i].frq = .5*(XD[i].frq + YC[i].frq);
			}
		}
	else if( M <= N )
		{
		for(int j=1; j<M; j++)
			YC[j] = XD[j];
		if(N < 125)
			for(int k=M; k<N; k++)
				YC.push_back(XD[k]);
		}
	printf("combine %d\n", YC.size());
}


float glissdetector(const deque <note> &DQ)
{
	float gl = 0, df1, df2=0;
	const int N = DQ.size();
	int s = 0;
	
	for(int i=1; i<N; i++)
		{
		df1 = DQ[i].frq - DQ[i-1].frq;
		if( (df1 > 0 && df2 > 0) ||
			(df1 < 0 && df2 < 0) )
		++s;
		df2 = df1;
		}

	gl = (float) s/N;
	return gl;
}






int main()
{


	deque <note> PD, GB;
	make_seq(PD, 12);
	make_seq(GB, 15);

	muddify(PD);
	permutate(PD, 0);
	muddify(PD);
	muddify(PD);
	permutate(PD, 1);
	muddify(PD);

	const int sr = 24000;
	chip CP(sr);
	
	SNDFILE *out;
	SF_INFO head;
	
	head.samplerate = sr;
	head.channels = 1;
	head.format = (SF_FORMAT_AU | SF_FORMAT_PCM_S8);
	
	out = sf_open("chiptuner_5.au", SFM_WRITE, &head);
	
	
	float x, gliss=1;
	const int L = sr/2;		// max - dur
	const unsigned int N = 300 ;
	float smp[L];
	float ff, frq, amp, A;
	lopass LP(sr, 50), SL(sr, 20);
	dc_block DC;
	
	srand(time(0));
	FILE *sco = fopen("chip5.sco", "w");
	fprintf(sco, "f1 0 4096 10 1\n\n");
	
	
	for (unsigned long int i=0, k, tt=0; i<N; i++)
		{
		k = rand() % 8;
		switch (k)
			{
			case 0:
				{	muddify(PD); break;	}
			case 1:
				{	permutate(PD, gliss); break; }
			case 2:
				{	conk(PD); break; }
			case 3:
				{	sift(PD); break; }
			case 4:
				{	shrimp(PD); break; }
			case 5: 
				{	tolchok(PD); break;	}
			case 6:
				{	combine(GB, PD); break; }
			default:
				havoc(PD);
				store(PD, GB);
			}

		gliss = glissdetector(PD);
		printf("\tgliss = %.4f\n", gliss);

		score_event(PD, k, sco, (float) tt/sr );

		for(unsigned int m=0, D; m<PD.size(); m++)
			{
			D = sr*PD[m].dur;
			if(D >= L) D = sr/2.25;
			ff = PD[m].frq;
			amp = PD[m].amp;
			tt += D;	// total time, #smp
			
			for(unsigned int j=0; j<D; j++)
				{
				A = SL.filter(amp);
				frq = LP.filter(ff);
				x = A*CP.play(ff, A);
				smp[j] = tanh(DC.filter(x));
				}
			sf_write_float(out, smp, D);
			}
		//
		}

	sf_close(out);
	fprintf(sco, "\ne\n");
	fclose(sco);
	
	return 0;
}

