#ifndef WAVESET_H
#define WAVESET_H

#include <cmath>

/*
	waveset.h 
	use for such things as waveset omission

	(c) Risto Holopainen 2014
	
	Copying and distribution of this file, with or without modification,
	are permitted in any medium without royalty provided the copyright
	notice and this notice are preserved.  This file is offered as-is,
	without any warranty.	

*/


class sieve
{
	public:
		sieve(int=2, int=1);				// P/Q: keep Q out of P cycles
		~sieve();
		double operator()(const double);	// zeros every other period
		double zero(const double);			// d:o, windowed
		bool drop(const double);			// drop cycles
		
	private:		
		inline bool is_switchpoint(double x);
		int P, Q;
		int cc;	// crosscount
		double xo;
		double y;
		double yo;
		int sz;
		int t;
		double *ew;
		double *qw;
};

sieve::sieve(int p, int q)
{
	P = p; Q = q;
	xo=0;	
	y=0;
	yo=0;
	cc=0;
	sz=60;
	ew = new double[sz];
	qw = new double[sz];
	t=0;	// time index for exp. window
	for(int i=0; i<sz; i++)
		{
		ew[i] = 0.5*(1.0 + cos(M_PI*i/sz));
		qw[i] = sqrt(ew[i]);
		}
}

sieve::~sieve()
{
  delete ew;
	delete qw;	
}

inline
bool sieve::is_switchpoint(double x)
{
  if(x * xo < 0) 		/* 0-crossing */
		++cc;

	xo = x;	
	bool is_crosspoint = ((cc/P) % P < Q);
  return is_crosspoint;
}



double sieve::zero(const double x)
{
	// smoothed version

	static bool reset_t = true;
	double b;
	bool sw_point = is_switchpoint(x);
		
	if(sw_point)	// signal approaching 0
		{
		if(reset_t == false)
			{
			t = 0;
			reset_t = true;
			}
		if(t < sz)		// forward exp-env
			y = ew[t++]*x;	
		else if(t >= sz)
			y = 0;		
		}

  else if(!sw_point)	//
  	{
		if(reset_t = true)
			{
	  	t = 0;
	  	reset_t = false;
			}
		if(t++ < sz)	// backwards exp.env
			b = qw[t];
		else if(t >= sz)
			b *= 0.875;
			
		y = (1.0-b)*x + b*yo;
		}

	yo = y;
  return y;
}	


double sieve::operator()(const double x)
{
  y = x;
	if(is_switchpoint(x))
		y = 0;
  
  return y;
}	


bool sieve::drop(const double x)
{
	return is_switchpoint(x);
}

#endif
